#!/usr/bin/env python3
#
# Copyright 2012 Ettus Research LLC
# Copyright 2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import subprocess, time
from optparse import OptionParser
from string import split
import sys
import os

try:
    from gnuradio.eng_option import eng_option
except:
    eng_option = None

def launch_test(args="", rate=None, spb=None, spp=0, prefix="", suffix="", extra=[], verbose=False, title=None):
    real = os.path.realpath(__file__)
    basedir = os.path.dirname(real)
    responder = [
        os.path.join(basedir, "responder")
    ]

    if args is not None and len(args) > 0:
        responder += ["--args=" + args]
    if rate is not None and rate > 0:
        responder += ["--rate=%f" % (rate)]
    if spb is not None and spb > 0:
        responder += ["--spb=%d" % (spb)]
    if spp is not None and spp > 0:
        responder += ["--spp=%d" % (spp)]
    if prefix is not None and len(prefix) > 0:
        responder += ["--stats-file-prefix=" + prefix]
    if suffix is not None and len(suffix) > 0:
        responder += ["--stats-file-suffix=" + suffix]
    if extra is not None:
        responder += extra
    if title is not None and len(title) > 0:
        responder += ["--title=\"" + title + "\""]
    if verbose:
        print "==> Executing:", " ".join(responder)
    try:
        responder += ["--log-file"] # This will produce another output file with logs
        responder += ["--combine-eob"]
        p = subprocess.Popen(responder)
        res = p.wait() # make sure subprocess finishes
    except KeyboardInterrupt:
        res = p.wait() # even in CTRL+C case wait till subprocess finishes
        print "==> Caught CTRL+C"
        return None

    return res

# These return codes should match the C++ return codes
class ReturnCode:
    RETCODE_OK = 0
    RETCODE_BAD_ARGS = -1
    RETCODE_RUNTIME_ERROR = -2
    RETCODE_UNKNOWN_EXCEPTION = -3
    RETCODE_RECEIVE_TIMEOUT = -4
    RETCODE_RECEIVE_FAILED = -5
    RETCODE_MANUAL_ABORT = -6
    RETCODE_BAD_PACKET = -7
    RETCODE_OVERFLOW = -8


def get_initialized_OptionParser():
    def_rates = ".25 1 4 8 25"
    usage = "%prog: [options] -- [extra arguments]"
    opt_kwds = {}
    if eng_option: opt_kwds['option_class'] = eng_option
    parser = OptionParser(usage=usage, **opt_kwds)

    parser.add_option("", "--rates", type="string", help="sample rates (Msps) [default: %default]", default=def_rates)
    parser.add_option("", "--spbs", type="string", help="samples per block [default: %default]",
                      default="32 64 256 1024")
    parser.add_option("", "--spps", type="string", help="samples per packet (0: driver default) [default: %default]",
                      default="0 64 128 256 512")
    parser.add_option("", "--args", type="string", help="UHD device arguments [default: %default]", default=None)
    parser.add_option("", "--prefix", type="string", help="Stats filename prefix [default: %default]", default=None)
    parser.add_option("", "--suffix", type="string", help="Stats filename suffix [default: %default]", default=None)
    parser.add_option("", "--pause", action="store_true", help="pause between tests [default=%default]", default=False)
    parser.add_option("", "--interactive", action="store_true", help="enable prompts within test [default=%default]",
                      default=False)
    parser.add_option("", "--wait", type="float", help="time to wait between tests (seconds) [default=%default]",
                      default=0.0)
    parser.add_option("", "--abort", action="store_true", help="abort on error [default=%default]", default=False)
    parser.add_option("", "--verbose", action="store_true", help="be verbose [default=%default]", default=False)
    parser.add_option("", "--title", type="string", help="test title [default: %default]", default=None)

    return parser


def set_gen_prefix(prefix, save_dir):
    if not save_dir[-1] == "/":
        save_dir = save_dir + "/"

    if prefix == None:
        if os.path.exists(save_dir) is not True:
            os.makedirs(save_dir)
        prefix = save_dir
    return prefix


def get_extra_args(options, args):
    extra_args = {
    "adjust-simulation-rate": None,
    "time-mul": "1e6",
    "test-success": 5,
    "simulate": 1000,
    "iterations": 1000,
    "delay-min": "50e-6",
    "delay-max": "5e-3",
    "delay-step": "50e-6",
    }

    if options.interactive is not True:
        extra_args["batch-mode"] = None
    if options.pause is True:
        extra_args["pause"] = None

    for arg in args:
        if len(arg) > 2 and arg[0:2] == "--":
            arg = arg[2:]
        idx = arg.find('=')
        if idx == -1:
            extra_args[arg] = None
        else:
            extra_args[arg[0:idx]] = arg[idx + 1:]

    def _format_arg(d, k):
        a = "--" + str(k)
        if d[k] is not None:
            a += "=" + str(d[k])
        return a

    extra = map(lambda x: _format_arg(extra_args, x), extra_args)

    print "\n".join(map(lambda x: str(x) + " = " + str(extra_args[x]), extra_args.keys()))

    return extra


def wait_for_keyboard():
    try:
        print "\nPress ENTER to start..."
        raw_input()
        return ReturnCode.RETCODE_OK
    except KeyboardInterrupt:
        print "Aborted"
        return ReturnCode.RETCODE_MANUAL_ABORT


def main():
    parser = get_initialized_OptionParser()
    (options, args) = parser.parse_args()

    save_dir = "results"
    options.prefix = set_gen_prefix(options.prefix, save_dir)
    extra = get_extra_args(options, args)

    rates = map(lambda x: float(x) * 1e6, split(options.rates))
    spbs = map(int, split(options.spbs))
    spps = map(int, split(options.spps))
    total = len(rates) * len(spbs) * len(spps)

    title = options.title or ""
    if len(title) >= 2 and title[0] == "\"" and title[-1] == "\"":
        title = title[1:-1]

    count = 0
    results = {}

    try:
        for rate in rates:
            results_rates = results[rate] = {}
            for spb in spbs:
                results_spbs = results_rates[spb] = {}
                for spp in spps:
                    if count > 0:
                        if options.pause:
                            print "Press ENTER to begin next test..."
                            raw_input()
                        elif options.wait > 0:
                            time.sleep(options.wait)
                    title = "Test #%d of %d (%d%% complete, %d to go)" % (
                        count + 1, total, int(100 * count / total), total - count - 1)
                    res = launch_test(options.args, rate, spb, spp, options.prefix, options.suffix, extra,
                                      options.verbose, title)
                    sys.stdout.flush()
                    count += 1
                    # Break out of loop. Exception thrown if Ctrl + C was pressed.
                    if res is None:
                        raise Exception
                    results_spbs[spp] = res
                    if res < 0 and (res == ReturnCode.RETCODE_MANUAL_ABORT or options.abort):
                        raise Exception
    except:
        pass

    for rate in results.keys():
        results_rates = results[rate]
        for spb in results_rates.keys():
            results_spbs = results_rates[spb]
            for spp in results_spbs.keys():
                res = results_spbs[spp]
                print res, ":", rate, spb, spp
    print "Tests finished"
    return 0


if __name__ == '__main__':
    main()



#
# Copyright 2024 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
rfnoc-newmod: Example module for Python support of an RFNoC OOT Module
"""

# Import all bindings from C++
from . import rfnoc_newmod_python as lib

# In UHD, we use CamelCase for names in Python, so we'll do the same here
#MyBlockControl = lib.my_block_control


#!/usr/bin/env python3
#
# Copyright 2010-2011 Ettus Research LLC
# Copyright 2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
The usrp2 recovery app:

When the usrp2 has an unknown or bad ip address in its eeprom,
it may not be possible to communicate with the usrp2 over ip/udp.

This app will send a raw ethernet packet to bypass the ip layer.
The packet will contain a known ip address to burn into eeprom.
Because the recovery packet is sent with a broadcast mac address,
only one usrp2 should be present on the interface upon execution.

This app requires super-user privileges and only works on linux.
"""

import socket
import struct
import argparse

BCAST_MAC_ADDR = 'ff:ff:ff:ff:ff:ff'
RECOVERY_ETHERTYPE = 0xbeee
IP_RECOVERY_CODE = b'addr'

def mac_addr_repr_to_binary_string(mac_addr):
    """
    Convert string representation of MAC address (xx:yy:zz:aa:bb:cc) into a
    binary string.
    """
    return bytes(bytearray([int(x, 16) for x in mac_addr.split(':')]))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(usage='usage: %(prog)s [options]\n'+__doc__)
    parser.add_argument(
        '--ifc',
        help='ethernet interface name [default=%(default)s]',
        default='eth0')
    parser.add_argument(
        '--new-ip',
        help='ip address to set [default=%(default)s]',
        default='192.168.10.2')
    args = parser.parse_args()

    #create the raw socket
    print("Opening raw socket on interface:", args.ifc)
    soc = socket.socket(socket.PF_PACKET, socket.SOCK_RAW)
    soc.bind((args.ifc, RECOVERY_ETHERTYPE))

    #create the recovery packet
    print("Loading packet with ip address:", args.new_ip)
    packet = struct.pack(
        '!6s6sH4s4s',
        mac_addr_repr_to_binary_string(BCAST_MAC_ADDR),
        mac_addr_repr_to_binary_string(BCAST_MAC_ADDR),
        RECOVERY_ETHERTYPE,
        IP_RECOVERY_CODE,
        socket.inet_aton(args.new_ip),
    )

    print("Sending packet ({} bytes)".format(len(packet)))
    soc.send(packet)
    print("Done")



#!/usr/bin/env python3
#
# Copyright 2012 Ettus Research LLC
# Copyright 2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import sys, re
from optparse import OptionParser

import matplotlib.pyplot as plt
import matplotlib.font_manager
import numpy as np

try:
    from gnuradio.eng_option import eng_option
except:
    eng_option = None

_units = [
    (3, "k"),
    (6, "M"),
    (9, "G")
]

def _format_rate(rate):
    for (u1, s1), (u2, s2) in zip(_units, _units[1:]):
        n = pow(10, u1)
        if rate >= n and rate < pow(10, u2):
            r = rate % n
            if r > 0:
                return str(1.0 * rate / n) + " " + s1
            else:
                return str(rate / n) + " " + s1
    return str(rate) + " "


def _sort(series, keys):
    if len(keys) == 0:
        return []
    key = keys[0]
    rev = False
    if key[0] == '-':
        key = key[1:]
        rev = True
    l = []
    for s in series:
        if s[key] not in l:
            l += [s[key]]
    l.sort()
    if rev:
        l.reverse()
    return [(key, l)] + _sort(series, keys[1:])


def _order(series, sort_list):
    if len(sort_list) == 0:
        return series
    (sort_key, sort_key_list) = sort_list[0]
    if len(sort_key_list) == 0:
        return []
    #print sort_key, sort_key_list
    l = []
    for s in series:
        if s[sort_key] == sort_key_list[0]:
            l += [s]
    #print l
    return _order(l, sort_list[1:]) + _order(series, [(sort_list[0][0], sort_list[0][1][1:])] + sort_list[1:])


def get_option_parser():
    usage = "%prog: [options]"
    opt_kwds = {}
    if eng_option: opt_kwds['option_class'] = eng_option
    parser = OptionParser(usage=usage, **opt_kwds)

    parser.add_option("", "--id", type="string", help="device ID [default: %default]", default=None)
    parser.add_option("", "--sort", type="string", help="sort order [default: %default]", default="rate -spb -spp")
    parser.add_option("", "--output", type="string", help="output file [default: %default]", default=None)
    parser.add_option("", "--output-type", type="string", help="output file type [default: %default]", default="pdf")
    parser.add_option("", "--output-size", type="string", help="output file size [default: %default pixels]",
                      default="1600,900")
    parser.add_option("", "--xrange", type="float", help="X range [default: %default]", default=None)
    parser.add_option("", "--title", type="string", help="additional title [default: %default]", default=None)
    parser.add_option("", "--legend", type="string", help="legend position [default: %default]", default="lower right")
    parser.add_option("", "--diff", action="store_true", help="compare results instead of just plotting them", default=None)
    return parser


def get_sorted_series(args, options):
    series = []

    if len(args) > 0:
        with open(args[0]) as f:
            lines = f.readlines()
    else:
        lines = sys.stdin.readlines()
    if lines is None or len(lines) == 0:
        return

    for line in lines:
        line = line.strip()
        if len(line) == 0:
            continue
        x = {'file': line}
        idx2 = 0
        idx = line.find("latency-stats")
        if idx > 0:
            x['prefix'] = line[0:idx]
        idx = line.find(".id_")
        if idx > -1:
            idx += 4
            idx2 = line.find("-", idx)
            x['id'] = line[idx:idx2]
            if options.id is None:
                options.id = x['id']
            elif options.id != x['id']:
                print "Different IDs:", options.id, x['id']
        idx = line.find("-rate_")
        if idx > -1:
            idx += 6
            idx2 = line.find("-", idx)
            x['rate'] = int(line[idx:idx2])
        idx = line.find("-spb_")
        if idx > -1:
            idx += 5
            idx2 = line.find("-", idx)
            x['spb'] = int(line[idx:idx2])
        idx = line.find("-spp_")
        if idx > -1:
            idx += 5
            #idx2 = line.find(".", idx)
            idx2 = re.search("\D", line[idx:])
            if idx2:
                idx2 = idx + idx2.start()
            else:
                idx2 = -1
            x['spp'] = int(line[idx:idx2])
        idx = line.rfind(".")
        if idx > -1 and idx >= idx2:
            idx2 = re.search("\d", line[::-1][len(line) - idx:])
            if idx2 and (idx2.start() > 0):
                idx2 = idx2.start()
                x['suffix'] = line[::-1][len(line) - idx:][0:idx2][::-1]
        print x
        series += [x]

    sort_keys = options.sort.split()
    print sort_keys
    sorted_key_list = _sort(series, sort_keys)
    print sorted_key_list
    series = _order(series, sorted_key_list)

    return series


def main():
    # Create object with all valid options
    parser = get_option_parser()

    # Read in given command line options and arguments
    (options, args) = parser.parse_args()


    # series contains path and attributes for all data sets given by args.
    series = get_sorted_series(args, options)

    # Read in actual data sets from file
    data = read_series_data(series)

    if options.diff:
        data = calculate_data_diff(data)


    # Get all the wanted properties for this plot
    plt_props = get_plt_props(options)
    print plt_props

    mpl_plot(data, plt_props)

    return 0


def read_series_data(series):
    if series is None: return []
    result = []
    for s in series:
        data = {}
        [data_x, data_y] = np.loadtxt(s['file'], delimiter=" ", unpack=True)
        data['x'] = data_x
        data['y'] = data_y
        data['metadata'] = s
        result.append(data)
    return result


def find_values(data, key):
    result = []
    for d in data:
        val = d['metadata'][key]
        if not val in result:
            result.append(val)
    return result


def find_match(data, key, val):
    result = []
    for d in data:
        meta = d['metadata']
        if meta[key] == val:
            result.append(d)
    return result

def get_data_diff(data):
    if not data:
        return data # just return. User didn't input any data.
    if len(data) < 2:
        return data[0] # Single data set. Can't calculate a diff.

    print "diff %d: rate %s, spb %s, spp %s" % (len(data), data[0]['metadata']['rate'], data[0]['metadata']['spb'], data[0]['metadata']['spp'])

    data = align_data(data)

    min_len = len(data[0]['x'])
    for d in data:
        min_len = min(min_len, len(d['x']))

    metadiff = ""
    for d in data:
        m = d['metadata']['prefix']
        for r in "/._":
            m = m.replace(r, "")
        metadiff += m + "-"

    xd = data[0]['x'][0:min_len]
    yd = data[0]['y'][0:min_len]
    meta = data[0]['metadata']
    meta['diff'] = metadiff
    other = data[1:]
    for d in other:
        y = d['y']
        for i in range(len(yd)):
            yd[i] -= y[i]

    result = {}
    result['x'] = xd
    result['y'] = yd
    result['metadata'] = meta
    return result

def align_data(data):
    x_start = 0
    for d in data:
        x_start = max(x_start, d['x'][0])

    for i in range(len(data)):
        s = np.where(data[i]['x'] == x_start)[0]
        data[i]['x'] = data[i]['x'][s:]
        data[i]['y'] = data[i]['y'][s:]

    return data


def calculate_data_diff(data):
    spps = find_values(data, "spp")
    spbs = find_values(data, "spb")
    rates = find_values(data, "rate")
    print spps, "\t", spbs, "\t", rates
    result = []
    for rate in rates:
        rd = find_match(data, "rate", rate)
        for spb in spbs:
            bd = find_match(rd, "spb", spb)
            for spp in spps:
                pd = find_match(bd, "spp", spp)
                if len(pd) > 0:
                    result.append(get_data_diff(pd))

    return result


def get_plt_props(options):
    plt_props = {}
    plt_out = None
    if options.output is not None:
        try:
            idx = options.output_size.find(",")
            x = int(options.output_size[0:idx])
            y = int(options.output_size[idx + 1:])
            plt_out = {'name': options.output,
                       'type': options.output_type,
                       'size': [x, y]}
        except:
            plt_out = None

    plt_props['output'] = plt_out

    if not options.id: options.id = "no data"
    plt_title = "Latency (" + options.id + ")"
    if options.title is not None and len(options.title) > 0:
        plt_title += " - " + options.title
    plt_props['title'] = plt_title

    plt_props['xlabel'] = "Latency (us)"
    plt_props['ylabel'] = "Normalised success of on-time burst transmission"

    plt_legend_loc = None
    if options.legend is not None:
        plt_legend_loc = options.legend
    plt_props['legend'] = plt_legend_loc

    plt_xrange = None
    if options.xrange is not None:
        plt_xrange = [0, options.xrange]
    plt_props['xrange'] = plt_xrange
    return plt_props


def mpl_plot(data, props):
    plt_out = props['output']
    plt_title = props['title']
    plt_xlabel = props['xlabel']
    plt_ylabel = props['ylabel']
    plt_legend_loc = props['legend']
    plt_xrange = props['xrange']

    markers = ['.', ',', 'o', 'v', '^', '<', '>', '1', '2', '3', '4', '8',
               's', 'p', '*', 'h', 'H', '+', 'D', 'd', '|', '_']
    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
    midx = 0

    # plot available data.
    mylegend = []
    for d in data:
        mylegend.append(get_legend_str(d['metadata']))
        plt.plot(d['x'], d['y'], marker=markers[midx], markerfacecolor=None)
        midx = (midx + 1) % len(markers)

    # Set all plot properties
    plt.title(plt_title)
    plt.xlabel(plt_xlabel)
    plt.ylabel(plt_ylabel)
    plt.grid()
    fontP = matplotlib.font_manager.FontProperties()
    fontP.set_size('x-small')
    plt.legend(mylegend, loc=plt_legend_loc, prop=fontP, ncol=2)
    if plt_xrange is not None:
        plt.xlim(plt_xrange)

    # Save plot to file, if option is given.
    if plt_out is not None:
        fig = plt.gcf() # get current figure
        dpi = 100.0 # Could be any value. It exists to convert the input in pixels to inches/dpi.
        figsize = (plt_out['size'][0] / dpi, plt_out['size'][1] / dpi) # calculate figsize in inches
        fig.set_size_inches(figsize)
        name = plt_out['name'] + "." + plt_out['type']
        plt.savefig(name, dpi=dpi, bbox_inches='tight')

    plt.show()


def get_legend_str(meta):
    lt = ""
    if meta.has_key('diff') and meta['diff']:
        lt += meta['diff'] + " "
    lt += "%ssps, SPB %d, SPP %d" % (_format_rate(meta['rate']), meta['spb'], meta['spp'])
    return lt


if __name__ == '__main__':
    main()
